/**
 * Menyskript för Tillsynsmanualen.
 *
 * Huvudenyn består av en menyrad med undermenyer.
 *
 * Innehållsmenyn, som finns på vissa sidor, är en meny. I smala vyer visas en
 * menyknapp som visar och döljer menyn.
 *
 * Menyerna använder flytande tangentbordsfokus, vilket innebär att bara ett
 * element i en grupp är fokuserbart med Tabb. Navigering inom gruppen sker
 * sedan med piltangenter (med mera), och flyttar då flytande fokus till ett
 * nytt element. Det här innebär att användaren slipper tabba sig igenom alla
 * menyalternativ, då hela menyn utgör bara ett tabbstopp.
 *
 * @see https://www.w3.org/TR/wai-aria-practices-1.1/#menu
 * @see https://www.w3.org/TR/wai-aria-practices-1.1/#menubutton
 *
 * Klasser
 * -------
 * Menu             Vertikal meny.
 * MenuBar          Horisontell menyrad.
 * MenuButton       Menyknapp som visar/döljer en meny i smala vyer.
 * MenuController   Controllerklass för följsam formgivning.
 * MenuItem         Ett valbart aternativ i en meny.
 * TopLevelMenuItem Ett valbart alternativ på översta nivån.
 */
((win, doc) => {
    "use strict";


    /**
     * En komponent som erbjuder användaren en vertikal lista med val.
     *
     * En `Menu`-instans kan vara en undermeny i en menyrad, men kan också vara
     * en fristående vertikal meny, till exempel den innehållsmeny som finns på
     * bland annat granskningssidorna.
     *
     * @property {string}      id         UL-listans `id`-attribut.
     * @property {string}      labelledBy UL-listans `aria-labelledBy`-attribut.
     * @property {MenuButton?} parent     Eventuell ägare.
     */
    class Menu {

        /**
         * Konstrera en menyinstans.
         *
         * Sätt rollen `menu` på UL-listan som utgör menyn och koppla alla
         * element till menyn.
         *
         * @param {HTMLUListElement} menu  Menyelement (oordnad lista).
         * @param {MenuItem[]}       items Vektor med menyelement.
         */
        constructor (menu, items) {
            if (!(menu instanceof HTMLUListElement)) {
                throw new Error("'menu' argument must be a UL element.");
            }

            this._hasRovingFocus = false;
            this._menu           = menu;
            this._items          = items;
            this._parent         = null;

            // Sätt ett ID på menyelementet.
            if ("" === this._menu.id) {
                this._menu.id = "menu-" + Math.round(1000 * Math.random());
            }
            this._menu.setAttribute("role", "menu");
            this._items.forEach((item) => item.parent = this);
        }

        // Åtkomstmetoder.
        get id         ()   {return this._menu.id                          }
        get parent     ()   {return this._parent                           }
        set labelledBy (id) {this._menu.setAttribute("aria-labelledby", id)}
        set parent     (p)  {this._parent = p                              }

        /**
         * Stäng undermenyer till referenselementets syskon.
         *
         * @param {MenuItem} ref Referenselement.
         */
        closeSiblings (ref) {
            (this._items || []).forEach((item) => {
                if (item !== ref) {
                    item.close();
                }
            });
        }

        /**
         * Sätt tangentbordsfokus till första menyelementet.
         *
         * Om menyn har flytande fokus, sätt även det till första elementet.
         */
        focusFirstItem () {
            this._items[0].focus();
            this._hasRovingFocus && this.setRovingFocus(this._items[0]);
        }

        /**
         * Sätt fokus till nästa menyelement vars text börjar med det
         * angivna tecknet.
         *
         * @param {MenuItem} ref  Referenselement.
         * @param {string}   char Tecken.
         * @return {boolean}
         */
        focusItemByChar (ref, char) {
            let   handled = false;
            const labels  = this._items
                            .map((item) => item.label.charAt(0).toLowerCase());
            const index   = this._items.indexOf(ref);
            let   next    = labels.indexOf(char, index + 1);

            if (-1 === next) {
                next = labels.indexOf(char);
            }
            if (-1 !== next) {
                this._items[next].focus();
                this._hasRovingFocus && this.setRovingFocus(this._items[next]);
                handled = true;
            }

            return handled;
        }

        /**
         * Sätt tangentbordsfokus till sista menyelementet.
         *
         * Om menyn har flytande fokus, sätt även det till första elementet.
         */
        focusLastItem () {
            const index = this._items.length - 1;
            this._items[index].focus();
            this._hasRovingFocus && this.setRovingFocus(this._items[index]);
        }

        /**
         * Sätt tangentbordsfokus till nästa menyelement.
         *
         * Om referenselementet är det sista elementet i menyn, gå runt till
         * det första elementet.
         *
         * Om menyn har flytande fokus, sätt även det till första elementet.
         *
         * @param {MenuItem} ref Referenselement.
         */
        focusNextItem (ref) {
            let   index  = this._items.indexOf(ref);
            const length = this._items.length;
            this._items[++index % length].focus();
            this._hasRovingFocus
            && this.setRovingFocus(this._items[index % length]);
        }

        /**
         * Sätt tangentbordsfokus till föregående menyelement.
         *
         * Om referenselementet är det första elementet i menyn, gå runt till
         * det sista elementet.
         *
         * Om menyn har flytande fokus, sätt även det till första elementet.
         *
         * @param {MenuItem} ref Referenselement.
         */
        focusPreviousItem (ref) {
            let   index  = this._items.indexOf(ref);
            const length = this._items.length;
            this._items[(--index + length) % length].focus();
            this._hasRovingFocus
            && this.setRovingFocus(this._items[index % length]);
        }

        /**
         * Sätt flytande tangentbordsfokus till ett menyelement.
         *
         * @param {MenuItem} item Menyelementet som ska få flytande fokus.
         */
        setRovingFocus (item = null)  {
            this._hasRovingFocus = true;
            item = item || this._items[0];
            this.unsetRovingFocus();
            item.item.tabIndex = 0;
        }

        /**
         * Ta bort flytande tangentbordsfokus från alla menyelement.
         *
         * @param {MenuItem} item Menyelementet som ska få flytande fokus.
         */
        unsetRovingFocus ()  {
            this._items.forEach((item) => item.item.tabIndex = -1);
        }
    }


    /**
     * Horisontell menyrad som är synlig hela tiden.
     */
    class MenuBar {

        /**
         * Konstruera en menyradsinstans.
         *
         * Sätt rollen `menu` på UL-listan som utgör menyn och koppla alla
         * element till menyn. Sätt också flytande fokus till det senast
         * fokuserade elementet.
         *
         * @param {HTMLUListElement}   menuBar Menyradselement (oordnad lists).
         * @param {TopLevelMenuItem[]} items   Vektor med  menyelement.
         */
        constructor (menuBar, items) {
            if (!(menuBar instanceof HTMLUListElement)) {
                throw new Error("'menuBar' argument must be a UL element.");
            }

            this._menuBar = menuBar;
            this._items   = items;

            this._menuBar.setAttribute("role", "menubar");
            this._items.forEach((item) => item.parent = this);

            this.setRovingFocus(
                this._items
                .find((item) => "current" === item.item.parentNode.className)
            );
        }

        /**
         * Stäng eventuell öppen undermeny.
         *
         * @return {boolean}
         */
        close () {
            return this._items.some((item) => item.close());
        }

        /**
         * Stäng undermenyer till referenselementets syskon.
         *
         * @param {TopLevelMenuItem} ref Referenselement.
         */
        closeSiblings (ref) {
            (this._items || []).forEach((item) => {
                if (item !== ref) {
                    item.close();
                }
            });
        }

        /**
         * Ange om menyraden eller en undermeny innehåller angivet element.
         *
         * @param {MenuItem} element Element att kontrollera emot.
         * @return {boolean}
         */
        contains (element) {
            return this._menuBar.contains(element);
        }

        /**
         * Sätt tangentbordsfokus och flytande fokus till första menyelementet.
         */
        focusFirstItem () {
            this._items[0].focus();
            this.setRovingFocus(this._items[0]);
        }

        /**
         * Sätt fokus till nästa menyelement vars text börjar med det
         * angivna tecknet.
         *
         * @param {MenuItem} ref  Referenselement.
         * @param {string}   char Tecken.
         * @returns {boolean}
         */
        focusItemByChar (ref, char) {
            let   handled = false;
            const labels  = this._items
                            .map((item) => item.label.charAt(0).toLowerCase());
            const index   = this._items.indexOf(ref);
            let   next    = labels.indexOf(char, index + 1);

            if (-1 === next) {
                next = labels.indexOf(char);
            }
            if (-1 !== next) {
                this._items[next].focus();
                handled = true;
            }

            return handled;
        }

        /**
         * Sätt tangentbordsfokus och flytande fokus till sista menyelementet.
         */
        focusLastItem () {
            let index = this._items.length - 1;
            this._items[index].focus();
            this.setRovingFocus(this._items[index]);
        }

        /**
         * Sätt tangentbordsfokus och flytande fokus till nästa menyelement.
         *
         * Om referenselementet är det sista elementet i menyn, gå runt till
         * det första elementet.
         *
         * @param {MenuItem} ref    Referenselement.
         * @param {boolean}  [open] Öppna elementets undermeny, om satt.
         */
        focusNextItem (ref, open = false) {
            let   index  = this._items.indexOf(ref);
            const length = this._items.length;

            index = ++index % length;
            this._items[index].focus();
            this.setRovingFocus(this._items[index]);

            if (open) {
                this._items[index].open();
            }
        }

        /**
         * Sätt tangentbordsfokus och flytande fokus till föregående menyelement.
         *
         * Om referenselementet är det första elementet i menyn, gå runt till
         * det sista elementet.
         *
         * @param {MenuItem} ref    Referenselement.
         * @param {boolean}  [open] Öppna elementets undermeny, om satt.
         */
        focusPreviousItem (ref, open = false) {
            let   index  = this._items.indexOf(ref);
            const length = this._items.length;

            index = (--index + length) % length;
            this._items[index].focus();
            this.setRovingFocus(this._items[index]);

            if (open) {
                this._items[index].open();
            }
        }

        /**
         * Sätt flytande tangentbordsfokus till ett menyradselement.
         *
         * @param {TopLevelMenuItem} item Elementet som ska få flytande fokus.
         */
        setRovingFocus (item)  {
            this._items.forEach((item) => item.item.tabIndex = -1);
            item.item.tabIndex = 0;
        }
    }


    /**
     * Knapp som visar/döljer en meny.
     *
     * @property {HTMLButtonElement} button Knappelement.
     * @property {string}            id     Knappens `id`-attributvärde.
     * @property {Menu}              menu   Tillhörande meny.
     */
    class MenuButton {

        /**
         * Konstruera en menyknappsinstans.
         *
         * Skapa ett `button`-element och koppla händelselyssnare.
         *
         * @param {HTMLElement} child Element som ska ersättas av knappen.
         * @param {string}      label Knapptext.
         * @param {string}      id    Knapp-ID.
         * @param {Menu}        menu  Menyinstans.
         */
        constructor (child, label, id, menu) {
            this._menu        = menu;
            this._menu.parent = this;

            const ul  = child.nextElementSibling;
            const div = child.parentNode
                        .insertBefore(doc.createElement("div"), ul);
            div.appendChild(ul);

            this._button      = doc.createElement("button");
            this._button.type = "button";
            this._button.id   = `${menu.id}-button`;

            this._button.setAttribute("aria-haspopup", "true");
            this._button.setAttribute("aria-controls", this._menu.id);

            this._button.appendChild(doc.createElement("span"))
            .appendChild(doc.createTextNode(label));
            child.parentNode.replaceChild(this._button, child);

            this._button.addEventListener("click", (event) => {
                this.close() || this.open();
                event.preventDefault();
                event.stopPropagation();
            });

            this._button.addEventListener("keydown", (event) => {
                if (this.handleKey(event)) {
                    event.preventDefault();
                    event.stopPropagation();
                }
            });

            this._menu.setRovingFocus();
        }

        // Åtkomstmetoder.
        get button () { return this._button   }
        get id     () { return this._button.id}
        get menu   () { return this._menu     }

        /**
         * Dölj menyn om den är synlig.
         *
         * @return {boolean}
         */
        close () {
            if (this._button.hasAttribute("aria-expanded")) {
                this._button.removeAttribute("aria-expanded");
                this._menu.unsetRovingFocus();

                return true;
            }

            return false;
        }

        /**
         * Sätt tangentbordsfokus till menyknappen.
         */
        focus () {
            this._button.focus();
        }

        /**
         * Hantera tangenttryckshändelser.
         *
         * Enter, Mellanslag och Pil ned: öppna menyn och sätt fokus till första
         * elementet.
         * Pil upp: öppna menyn och sätt fokus till sista elementet.
         *
         * @param {KeyboardEvent} event Händelseinformation.
         * @return {boolean}
         */
        handleKey (event) {
            let handled = true;

            if (!(event.ctrlKey || event.ShiftKey || event.altKey
                  || event.metaKey)
            ) {
                switch (event.key) {
                    case "Enter":
                    case "Space":
                    case "ArrowDown":
                        this.open();
                        this._menu.focusFirstItem();
                        break;

                    case "ArrowUp":
                        this.open();
                        this._menu.focusLastItem();
                        break;

                    default:
                        handled = false;
                        break;
                }
            }

            return handled;
        }

        /**
         * Visa menyn om den är dold.
         *
         * @return {boolean}
         */
        open () {
            if (!this._button.hasAttribute("aria-expanded")) {
                this._button.setAttribute("aria-expanded", "true");

                return true;
            }

            return false;
        }

    }


    /**
     * Menycontroller för följsam formgivning.
     *
     * Aktiverar eller avaktiverar flytande fokus i innehållsmenyn, beroende på
     * om den är synlig eller inte.
     */
    class MenuController {

        /**
         * Konstruera en controllerinstans.
         *
         * Skapa en lyssnare för matchning av CSS-mediefrågor.
         *
         * @param {HTMLButtonElement} button  Menyknapp.
         */
        constructor (button) {
            this._active = null;
            this._button = button;

            // Skapa en matchning för mediefrågor och lägg till en lyssnare.
            let query = win.matchMedia("(min-width:50em)");
            query.addListener((query) => this.update(query));

            // Initiera hanteringen av flytande fokus.
            this.update(query);
        }

        /**
         * Aktivera eller avaktivera flytande fokus i innehållsmenyn, beroende
         * på vystorleken.
         *
         * @param {MediaQueryList} quey Mediefrågelista.
         */
        update (query) {
            if (query.matches) {
                this._button.menu.setRovingFocus();
            } else {
                this._button.menu.unsetRovingFocus();
            }
        }

    }


    /**
     * Ett val i en grupp av val som ingår i en meny.
     *
     * @property {HTMLElement} item  Underliggande HTML-element.
     * @property {string}      label Elementets textinnehåll.
     */
    class MenuItem {

        /**
         * Konstruera en menyelementsinstans.
         *
         * Sätt rollen `menuitem` på HTML-elementet. Om det har en undermeny,
         * sätt ytterligare ARIA-attribut och sätt även rollen `menu` på
         * UL-listan som innehåller undermenyn.
         *
         * @param {HTMLElement}      item   Element (länk, span eller strong).
         * @param {HTMLUListElement} menu   Menyelement (oordnad lista).
         * @param {Menu}             parent Överordnad Menu-instans.
         */
        constructor (item, menu = null, parent = null) {
            if (!item || !item.nodeName
                || -1 == ["A", "SPAN", "STRONG"].indexOf(item.nodeName)
            ) {
                throw new Error(
                    `'item' argument must be an A (anchor), SPAN, or
STRONG-element (${item.nodeName} given.)`
                );
            }

            if (menu && !(menu instanceof HTMLUListElement)) {
                throw new Error("'menu' argument must be a UL element.");
            }

            this._item   = item;
            this._menu   = menu;
            this._parent = parent;
            this._top    = false;

            // Ta bort semantisk betydelse från överordnat LI-element.
            if (this._item.parentNode
                && "LI" === this._item.parentNode.nodeName
            ) {
                this._item.parentNode.setAttribute("role", "none");
            }

            // Sätt rollen `menuitem` på menyelementet.
            this._item.setAttribute("role", "menuitem");
            this._item.tabIndex = -1;

            // Hantera tangentbordshändelser.
            this._item.addEventListener("keydown", (event) => {
                if (this.handleKey(event)) {
                    event.preventDefault();
                    event.stopPropagation();
                }
            });

            if (this._menu) {
                // Sätt ARIA-attribut på menyelementet.
                this._item.setAttribute("aria-haspopup", "true");
                this._item.setAttribute("aria-expanded", "false");
                this._item.href = "#";

                // Sättt roll och etikett för menyn.
                this._menu.setAttribute("role",       "menu");
                this._menu.setAttribute("aria-label", this._item.textContent);

                // Händelselyssnare för elementet.
                this._item.addEventListener("click", (event) => {
                    if (this.close() || this.open()) {
                        event.preventDefault();
                        event.stopPropagation();
                    }
                });

                // Skapa MenuItem-instanser för varje element i undermenyn.
                this._subItems = Array.prototype.slice.call(this._menu.children)
                                 .map((li) => {
                                     const a = li.firstElementChild;
                                     return new MenuItem(
                                         a,
                                         a.nextElementSibling,
                                         this
                                     );
                });
            }
        }

        // Åtkomstmetoder.
        get item  () {return this._item            }
        get label () {return this._item.textContent}

        /**
         * Stäng eventuell undermeny till elementet.
         *
         * @return {boolean}
         */
        close () {
            if ("true" === this._item.getAttribute("aria-expanded")) {
                this._item.setAttribute("aria-expanded", "false");

                return true;
            }

            return false;
        }

        /**
         * Stäng undermenyer till referenselementets syskon.
         *
         * @param {MenuItem} ref Referenselement.
         */
        closeSiblings (ref) {
            (this._subItems || []).forEach((item) => {
                if (item !== ref) {
                    item.close();
                }
            });
        }

        /**
         * Sättt tangentbordsfokus till menyelementet.
         */
        focus () {
            this._item.focus();
        }

        /**
         * Sätt tangentbordsfokus till första menyelementet.
         */
        focusFirstItem () {
            if (this._subItems) {
                this.open();
                this._subItems[0].focus();
            }
        }

        /**
         * Sätt fokus till nästa menyelement vars text börjar med det
         * angivna tecknet.
         *
         * @param {MenuItem} ref  Referenselement.
         * @param {string}   char Tecken.
         *
         * @return True om ett matchande element hittades, annars false.
         */
        focusItemByChar (ref, char) {
            let   handled = false;
            const labels  = this._subItems
                            .map((item) => item.label.charAt(0).toLowerCase());
            const index   = this._subItems.indexOf(ref);
            let   next    = labels.indexOf(char, index + 1);

            if (-1 === next) {
                next = labels.indexOf(char);
            }
            if (-1 !== next) {
                this._subItems[next].focus();
                handled = true;
            }

            return handled;
        }

        /**
         * Sätt tangentbordsfokus till sista menyelementet.
         */
        focusLastItem () {
            if (this._subItems) {
                this.open();
                this._subItems[this._subItems.length - 1].focus();
            }
        }

        /**
         * Sätt tangentbordsfokus till nästa menyelement.
         *
         * Om referenselementet är det sista elementet i menyn, gå runt till
         * det första elementet.
         *
         * @param {MenuItem} ref Referenselement.
         */
        focusNextItem (ref) {
            let   index  = this._subItems.indexOf(ref);
            const length = this._subItems.length;
            this._subItems[++index % length].focus();
        }

        /**
         * Sätt tangentbordsfokus till föregående menyelement.
         *
         * Om referenselementet är det första elementet i menyn, gå runt till
         * det sista elementet.
         *
         * @param {MenuItem} ref Referenselement.
         */
        focusPreviousItem (ref) {
            let   index  = this._subItems.indexOf(ref);
            const length = this._subItems.length;

            this._subItems[(--index + length) % length].focus();
        }

        /**
         * Hantera tangentbordshändelser.
         *
         * Enter, Blank: Öppna undermenyn och sätt fokus till första elementet;
         *               om undermeny saknas: aktivera elementet och stäng
         *               menyn.
         * Escape:       Stäng menyn och sätt fokus till överordnat element.
         * Home:         Sätt fokus till första underelementet.
         * End:          Sätt fokus till sista underelementet.
         * Pil upp:      Sätt fokus till föregående element (går runt).
         * Pil ned:      Sätt fokus till nästa element (går runt).
         * Pil vänster:  I en undermeny: stäng undermeny och sätt fokus till
         *               föräldern; i en toppmeny: stäng menyn, sätt fokus på
         *               föregående menyradselement och öppna dess undermeny.
         * Pil höger:    I en undermeny: öppna undermenyn och sätt fokus till
         *               första elementet; i en toppmeny: stäng menyn och
         *               överordnade menyer, sätt fokus till nästa
         *               menyradselement och öppna dess undermeny.
         * Övriga:       Sätt fokus till nästa element vars text börjar med
         *               detta tecken.
         *
         * @param {KeyboardEvent} event Händelseinformation.
         * @return {boolean}
         */
        handleKey (event) {
            let handled = false;

            if (!(event.ctrlKey || event.ShiftKey || event.altKey
                  || event.metaKey)
            ) {
                handled = true;

                switch (event.key) {
                    case "Enter":
                    case " ":
                        if (this._subItems) {
                            this.focusFirstItem();
                        } else {
                            this.close();
                            this._item.click();
                        }
                        break;

                    case "Escape":
                        this._parent.focus();
                        this._parent.close();
                        break;

                    case "Home":
                        this._parent.focusFirstItem();
                        break;

                    case "End":
                        this._parent.focusLastItem();
                        break;

                    case "ArrowUp":
                        this._parent.focusPreviousItem(this);
                        break;

                    case "ArrowDown":
                        this._parent.focusNextItem(this);
                        break;

                    case "ArrowLeft":
                        if (this._parent._top) {
                            this._parent.close();
                            this._parent._parent
                            .focusPreviousItem(this._parent, true);
                        } else {
                            this._parent.focus();
                            this._parent.close();
                        }
                        break;

                    case "ArrowRight":
                        if (this._subItems) {
                            this.focusFirstItem();
                        } else {
                            this._parent.handleRightKey(this);
                        }
                        break;

                    default:
                        handled = this._parent.focusItemByChar(
                            this,
                            event.key.toLowerCase()
                        );
                        break;
                }
            }

            return handled;
        }

        /**
         * Hantera högerpil i en undermeny.
         *
         * Stäng undermenyn och anropa rekursivt för att stänga den
         * överordnade menyn.
         *
         * @param {MenuItem} ref Referenselement.
         */
        handleRightKey (ref) {
            this.close();
            this._parent.handleRightKey(this);
        }

        /**
         * Öppna tillhörande undermeny.
         *
         * @return {boolean}
         */
        open () {
            if (this._menu && "false" === this._item.getAttribute("aria-expanded")) {
                this._parent.closeSiblings(this);
                this._item.setAttribute("aria-expanded", "true");

                return true;
            }

            return false;
        }
    }


    /**
     * Ett val i en grupp av val på högsta nivå som ingår i en meny eller
     * menyrad.
     *
     * @property {Menu|MenuBar} parent Ägare.
     */
    class TopLevelMenuItem extends MenuItem {

        /**
         * Konstruera en instans av ett menyalternativ på toppnivå.
         *
         * @param {HTMLAnchorElement} item Element.
         * @param {HTMLUListElement}  menu Menyelement (oordnad lista).
         */
        constructor (item, menu = null) {
            super(item, menu);
            this._top = true;

            // @todo Eliminera behov av info om ägare.
            this._item.addEventListener("focus", () => {
                if (this._parent instanceof MenuBar) {
                    this._parent.setRovingFocus(this);
                }
            });
        }

        // Åtkomstmetoder.
        set parent (parent) {this._parent = parent}

        /**
         * Hantera tangentbordshändelser.
         *
         * Enter, Blank, Pil ned: Öppna undermenyn och sätt fokus till första
         *                        elementet.
         * Escape:                Stäng undermenyn och sätt fokus till detta
         *                        element.
         * Pil upp:               Öppna undermenyn och sätt fokus till sista
         *                        elementet.
         * Pil vänster:           Stäng undermenyn och sätt fokus till
         *                        föregående menyradselement.
         * Pil höger:             Stäng undermenyn och sätt fokus till nästa
         *                        menyradselement.
         * Övriga:                Låt föräldraklassen hantera händelsen.
         *
         * @param {KeyboardEvent} event Händelseinformation.
         *
         * @return {boolean}
         */
        handleKey (event) {
            let handled = false;

            if (!(event.ctrlKey || event.ShiftKey || event.altKey
                  || event.metaKey)
            ) {
                handled = true;

                switch (event.key) {
                    case "Escape":
                        if (this._subItems) {
                            if (handled = this.close()) {
                                this.focus();
                            } else {
                                this._item.blur();
                            }
                        } else if (this._parent) {
                            this._parent.parent.close();
                            this._parent.parent.focus();
                        }
                        break;

                    case "ArrowUp":
                        if (this._subItems) {
                            this.focusLastItem();
                        } else {
                            this._parent.focusPreviousItem(this);
                        }
                        break;

                    case "ArrowDown":
                        if (this._subItems) {
                            this.focusFirstItem();
                        } else {
                            this._parent.focusNextItem(this);
                        }
                        break;

                    case "ArrowLeft":
                        this.close();
                        this._parent.focusPreviousItem(this);
                        break;

                    case "ArrowRight":
                        this.close();
                        this._parent.focusNextItem(this);
                        break;

                    default:
                        handled = super.handleKey(event);
                        break;
                }
            }

            return handled;
        }

        /**
         * Hantera vänsterpil i en undermeny till ett menyradselement.
         *
         * Stäng undermenyn, sätt fokus till föreående menyradselement och
         * öppna dess undermeny.
         *
         * @param {MenuItem} ref Referenselement.
         */
        handleLeftKey (ref) {
            this.close();
            this._parent.focusPreviousItem(ref, true);
        }

        /**
         * Hantera högerpil i en undermeny till ett menyradselement.
         *
         * Stäng undermeny, sätt fokus till nästa menyradselement och
         * öppna dess undermeny.
         *
         * @param {MenuItem} ref Referenselement.
         */
        handleRightKey (ref) {
            this.close();
            this._parent.focusNextItem(this, true);
        }
    }


    /*
     * Initiering.
     */

    // Kontrollera kompatibilitet.
    if (   "undefined" === typeof Array.prototype.slice
        || "undefined" === typeof Array.prototype.forEach
        || "undefined" === typeof Array.prototype.map
        || "undefined" === typeof win.matchMedia
        || "undefined" === typeof doc.querySelector
        || "undefined" === typeof doc.body.classList
    ) {
        return;
    }

    // Menyalternativ och menyrad.
    const items   = Array.prototype.slice
        .call(doc.querySelectorAll("header nav>ul>li>*:first-child"))
        .map((a) => new TopLevelMenuItem(a, a.nextElementSibling));
    const menuBar = new MenuBar(doc.querySelector("header nav>ul"), items);

    // Eventuell innehållsmeny och menyknapp.
    const nav = doc.querySelector("main:not(.glossary)>nav");
    if (nav) {
        const contentItems = Array.prototype.slice
            .call(doc.querySelectorAll("main>nav>ul>li>a"))
            .map((a) => new TopLevelMenuItem(a));

        const contentMenu = new Menu(
            doc.querySelector("main>nav>ul"),
            contentItems
        );

        const menuButton = new MenuButton(
            doc.querySelector("main>nav>a"),
            "Innehåll", 
            "content-menu-button",
            contentMenu
        );
        contentMenu.labelledBy = menuButton.id;

        nav.className = "menu";

        // Controller.
        new MenuController(menuButton);
        doc.body.addEventListener("click", () => menuButton.close());
    }

    // Klick eller fokus utanför någon öppen menu bör stänga menyn.
    doc.body.addEventListener("click", () => menuBar.close());
    doc.addEventListener("focus", (event) => {
        if (!menuBar.contains(event.target)) {
            menuBar.close();
        }
    }, true);

    // Sätt en klass på BODY för att tillämpa CSS-regler.
    doc.body.classList.add("js-menu");
})(window, document);
